package hpdos_rdma_offloaded.handler;

import java.io.IOException;
import org.apache.log4j.Logger;
import org.checkerframework.checker.units.qual.degrees;

import java.lang.System.Logger.Level;
import java.util.concurrent.CopyOnWriteArrayList;

import com.ibm.disni.util.NativeAffinity;

import hpdos_rdma_offloaded.protocol.InvalidationRequest;
import hpdos_rdma_offloaded.protocol.InvalidationResponse;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.rdma.DaRPCFutureM;
import hpdos_rdma_offloaded.rdma.DaRPCStreamM;

public class NetworkHandlerM implements Runnable{

    private CopyOnWriteArrayList<DaRPCStreamM<InvalidationRequest, InvalidationResponse>> invalidationStreams;
    private CopyOnWriteArrayList<DaRPCStreamM<Request, Response>> replicationStreams;
    final static Logger logger = Logger.getLogger(NetworkHandlerM.class);

    public NetworkHandlerM()
    {
        this.invalidationStreams = new CopyOnWriteArrayList<>();
        this.replicationStreams = new CopyOnWriteArrayList<>();
    }
    public void addInvalidationStream(DaRPCStreamM<InvalidationRequest, InvalidationResponse> stream){
        this.invalidationStreams.add(stream);
    }
    public void addReplicationStream(DaRPCStreamM<Request,Response> stream)
    {
        this.replicationStreams.add(stream);
    }
    void processCloseInvalidationStream(DaRPCStreamM<InvalidationRequest,InvalidationResponse> stream)
    {
        stream.completedList.forEach(future -> 
        {
            if(future.metadataResponse.invalidationTotal == future.metadataResponse.invalidationDone.incrementAndGet())
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex);   
                    }
                }
        });
        stream.endpoint.pendingFutures.forEach((ticket,future)->{
            if(future.metadataResponse.invalidationTotal == future.metadataResponse.invalidationDone.incrementAndGet())
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex);   
                    }
                }
        });
    }
    void processInvalidationStream(DaRPCStreamM<InvalidationRequest,InvalidationResponse> stream)
    {
        try
        {
            DaRPCFutureM<InvalidationRequest,InvalidationResponse> future = stream.processStream();
            if(future!=null)
            {
                //logger.debug("future"+future.metadataResponse.invalidationTotal+" "+future.metadataResponse.invalidationDone);
                //logger.debug(future.metadataResponse.state);
                if(future.metadataResponse.invalidationTotal == future.metadataResponse.invalidationDone.incrementAndGet())
                    if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                      future.metadataResponse.event.triggerResponse();           
            }
        }catch(Exception exx)
        {
            exx.printStackTrace();
        }
    }
    void processCloseReplicationStream(DaRPCStreamM<Request,Response> stream)
    {
        stream.completedList.forEach(future -> 
        {
            if(future.metadataResponse.replicationDone.incrementAndGet() == future.metadataResponse.replicatonTotal)
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())                
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex);   
                    }
                }
        });
        stream.endpoint.pendingFutures.forEach((ticket,future)->{
            if(future.metadataResponse.replicationDone.incrementAndGet() == future.metadataResponse.replicatonTotal)
                if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                {
                    try
                    {
                        future.metadataResponse.event.triggerResponse();
                    }catch(IOException ex)
                    {
                        logger.error( ex);   
                    }
                }
        });
    }
    void processReplicationStream(DaRPCStreamM<Request,Response> stream)
    {
        try
        {
            DaRPCFutureM<Request,Response> future = stream.processStream();
            if(future!=null)
            {
                if(future.metadataResponse.replicationDone.incrementAndGet() == future.metadataResponse.replicatonTotal)
                    if(Response.DONE == future.metadataResponse.state.incrementAndGet())
                      future.metadataResponse.event.triggerResponse();           
            }
        }catch(Exception exx)
        {
            exx.printStackTrace();
        }
    }
    public void run()
    {
        NativeAffinity.setAffinity(1<<1);
        while(true)
        {
            invalidationStreams.forEach(stream  ->
            {
                if( stream.endpoint.isClosed())
                {
                    processCloseInvalidationStream(stream);
                    invalidationStreams.remove(stream);
                }
                else
                {
                    processInvalidationStream(stream);
                }
            });
            replicationStreams.forEach(stream  ->
            {
                if( ! stream.endpoint.isClosed())
                {
                    processCloseReplicationStream(stream);
                    replicationStreams.remove(stream);
                }
                else
                {
                    processReplicationStream(stream);
                }
            });
        }
    }
    public void sendInvalidations(Request request,Response response)
    {
        InvalidationRequest inRequest =  new InvalidationRequest();
        //logger.debug("invdlidation send "+invalidationStreams.size());
        inRequest.setKey( request.getKey());
        response.invalidationTotal = invalidationStreams.size();

        for(var stream : invalidationStreams)
        {
            try
            {
                stream.request(inRequest, new InvalidationResponse(), response, true);
            }
            catch(IOException ex)
            {
                logger.error(ex);
            }
        }   

        if(invalidationStreams.size() == 0)
        {
            response.state.incrementAndGet();
        }
    }


    public void sendReplications(Request request,Response response)
    {
        logger.debug("replication send "+replicationStreams.size());
        /*Request repRequest =  new Request();
        repRequest.setRequestType(request.getRequestType());
        repRequest.setKey(request.getKey());

        if(request.getRequestType() == RequestType.PUT)
        {
            repRequest.setValue(request.getValue());
        }
        
        */
        response.replicatonTotal = replicationStreams.size();
        
        for(var stream : replicationStreams)
        {
            try
            {
                stream.request(request, new Response(), response, true);
            }
            catch(IOException ex)
            {
                logger.error(ex);
            }
        }  
        
        if(replicationStreams.size() == 0)
        {
            response.state.incrementAndGet();
        } 
    }
}
